import pytest
import asyncio
from unittest.mock import AsyncMock, patch

from pygryfsmart.api import GryfApi
from pygryfsmart.const import DriverFunctions

@pytest.mark.asyncio
async def test_init():
    """Test inicjalizacji klasy."""
    with patch("mypackage.gryf_communication.RS232Handler"), \
         patch("mypackage.gryf_communication.TCPClientHandler"), \
         patch("pygryfsmart.gryf_communication.Feedback"):
        
        api = GryfApi("192.168.1.100")
        assert api.feedback is not None
        assert api._writer is not None


@pytest.mark.asyncio
async def test_detect_connection_type():
    """Test wykrywania IP oraz portu szeregowego."""
    with patch("pygryfsmart.gryf_communication.RS232Handler"), \
         patch("pygryfsmart.gryf_communication.TCPClientHandler"):
        
        api_ip = GryfApi("192.168.1.100")
        assert isinstance(api_ip._writer, AsyncMock)  

        api_serial = GryfApi("/dev/ttyUSB0")
        assert isinstance(api_serial._writer, AsyncMock)  


@pytest.mark.asyncio
async def test_send_data():
    """Test wysyłania danych przez send_data()."""
    with patch("pygryfsmart.gryf_communication.RS232Handler") as mock_serial:
        mock_writer = mock_serial.return_value
        mock_writer.send_data = AsyncMock()

        api = GryfApi("/dev/ttyUSB0")
        await api.send_data("TEST_COMMAND")

        mock_writer.send_data.assert_called_once_with("TEST_COMMAND")


@pytest.mark.asyncio
async def test_subscribe_messages():
    """Test subskrypcji wiadomości."""
    with patch("mypackage.gryf_communication.RS232Handler"):
        api = GryfApi("/dev/ttyUSB0")

        def test_callback(data):
            return data

        api.subscribe_input_message(test_callback)
        api.subscribe_output_message(test_callback)

        assert test_callback in api._input_message_subscribers
        assert test_callback in api._output_message_subscribers


@pytest.mark.asyncio
async def test_start_stop_connection():
    """Test uruchamiania i zatrzymywania połączenia."""
    with patch("mypackage.gryf_communication.RS232Handler") as mock_serial:
        mock_writer = mock_serial.return_value
        mock_writer.open_connection = AsyncMock()
        mock_writer.close_connection = AsyncMock()

        api = GryfApi("/dev/ttyUSB0")

        await api.start_connection()
        assert api._connection_task is not None
        mock_writer.open_connection.assert_called_once()

        await api.stop_connection()
        assert api._connection_task is None
        mock_writer.close_connection.assert_called_once()


@pytest.mark.asyncio
async def test_available_driver():
    """Test dostępności drivera."""
    with patch("mypackage.gryf_communication.RS232Handler"):
        api = GryfApi("/dev/ttyUSB0")

        now = datetime.now().strftime("%H:%M")
        api.feedback.data = {DriverFunctions.PONG: {1: now}}

        assert api.available_driver(1) == True  
        assert api.available_driver(2) == False



from datetime import datetime

from .typing import GryfData

import logging
_LOGGER = logging.getLogger(__name__)

class Parser:

    _data: GryfData

    def __init__(self, feedback):
        self._feedback = feedback

        self._data = feedback._data

    async def parse_metod_1(self , parsed_states , line: str , function: str):
        if len(parsed_states) not in {7 , 9}:
            raise ValueError(f"Invalid number of arguments: {line}")

        id = int(parsed_states[0])
        if id not in self._data[function]:
            self._data[function][id] = {}

        for i in range(1, len(parsed_states)):
            if parsed_states[i] not in {"0" , "1"}:
                raise ValueError(f"Wrong parameter value: {line}")

            self._data[function][id][i] = int(parsed_states[i])                   
        try:
            await self._feedback.handle_subscribtion(function, id=id)
        except Exception as e:
            _LOGGER.error(f"Error subscriber {e}")

    async def parse_metod_2(self , parsed_states , line: str , function: str , prefix: int):
        if parsed_states[1] not in {"1" , "2" , "3" , "4" , "5" , "6" , "7" , "8"}:
            raise ValueError(f"Argument out of scope: {line}")

        pin = int(parsed_states[1])
        id = int(parsed_states[0])

        if id not in self._data[function]:
            self._data[function][id] = [0] * 20
        self._data[function][id][pin] = prefix
        try:
            await self._feedback.handle_subscribtion(function, id=id)
        except Exception as e:
            _LOGGER.error(f"Error subscriber {e}")
            
    async def parse_metod_3(self , parsed_states , line: str , function: str):
        if parsed_states[0] not in {"1" , "2" , "3" , "4" , "5" , "6" , "7" , "8"}:
            raise ValueError(f"Argument out of scope: {line}")

        pin = int(parsed_states[1])
        id = int(parsed_states[0])
        if id not in self._data[function]:
            self._data[function][id] = {}
        self._data[function][id][pin] = parsed_states[2]
        try:
            await self._feedback.handle_subscribtion(function, id)
        except Exception as e:
            _LOGGER.error(f"Error subscriber {e}")

    async def parse_cover(self , parsed_states , line: str , function: str):
        if len(parsed_states) != 5:
            raise ValueError(f"Invalid number of arguments: {line}")

        for i in range(1, len(parsed_states)):
            if parsed_states[i] not in {"0" , "1" , "2"}:
                raise ValueError(f"Wrong parameter value: {line}")

            pin = int(parsed_states[0])
            if pin not in self._data[function]:
                self._data[function][pin] = {}
            self._data[function][pin][i] = int(parsed_states[i])                   

        id = int(parsed_states[0])

        try:
            await self._feedback.handle_subscribtion(function, id)
        except Exception as e:
            _LOGGER.error(f"Error subscriber {e}")

    async def parse_temp(self , parsed_states , line: str):
        pin = int(parsed_states[1])
        id = int(parsed_states[0])
        if id not in self._data.temps:
            self._data.temps[id] = {}
        self._data.temps[id][pin] = float(f"{parsed_states[2]}.{parsed_states[3]}")
        try:
            await self._feedback.handle_temp_subscribtion(id , pin)
        except Exception as e:
            _LOGGER.error(f"Error subscriber {e}")

    async def parse_find(self , parsed_states):
        id = int(parsed_states[0])
        self._data.update_model(id, parsed_states[1], parsed_states[2])
        self._data.update_pong(int(parsed_states[0]))

    async def parse_pong(self , parsed_states):
        self._data.update_pong(int(parsed_states[0]))



import pytest
from unittest.mock import AsyncMock, patch
from pygryfsmart.rs232 import RS232Handler

@pytest.fixture
def rs232_handler():
    return RS232Handler(port="/dev/ttyUSB0", baudrate=115200)
pytest.mark.asyncio
@patch("serial_asyncio.open_serial_connection")
async def test_open_connection(mock_open_serial):
    mock_reader = AsyncMock()
    mock_writer = AsyncMock()
    mock_open_serial.return_value = (mock_reader, mock_writer)

    rs232_handler = RS232Handler(port="/dev/ttyUSB0", baudrate=115200)

    await rs232_handler.open_connection()

    mock_open_serial.assert_called_once_with(url="/dev/ttyUSB0", baudrate=115200)
    mock_writer = AsyncMock()
    rs232_handler.writer = mock_writer

    await rs232_handler.close_connection()

    mock_writer.close.assert_called_once()
    await mock_writer.wait_closed()

@pytest.mark.asyncio
async def test_send_data(rs232_handler):
    mock_writer = AsyncMock()
    rs232_handler.writer = mock_writer

    await rs232_handler.send_data("Test data")

    mock_writer.write.assert_called_once_with(b"Test data")
    await mock_writer.drain()

@pytest.mark.asyncio
async def test_read_data(rs232_handler):
    mock_reader = AsyncMock()
    mock_reader.readuntil = AsyncMock(return_value=b"Test data\n")
    rs232_handler.reader = mock_reader

    data = await rs232_handler.read_data()

    assert data == "Test data"
    mock_reader.readuntil.assert_called_once_with(b"\n")



import pytest
from unittest.mock import AsyncMock
from pygryfsmart.feedback import Feedback
from pygryfsmart.const import (
    COMMAND_FUNCTION_IN,
    COMMAND_FUNCTION_OUT,
    COMMAND_FUNCTION_PWM,
    COMMAND_FUNCTION_COVER,
    COMMAND_FUNCTION_FIND,
    COMMAND_FUNCTION_PONG,
    COMMAND_FUNCTION_PRESS_SHORT,
    COMMAND_FUNCTION_PRESS_LONG,
    COMMAND_FUNCTION_TEMP,
)


@pytest.fixture
def feedback():
    return Feedback()


@pytest.mark.asyncio
async def test_parse_metod_1(feedback):
    line = "IN=1,1,0,1,0,1,0,1;"
    parsed_states = ["1", "1", "0", "1", "0", "1", "0", "1" , "1"]
    await feedback._Feedback__parse_metod_1(parsed_states, line, COMMAND_FUNCTION_IN)

    assert feedback.data[COMMAND_FUNCTION_IN][1] == {1: 1, 2: 0, 3: 1, 4: 0, 5: 1, 6: 0, 7: 1, 8: 1}


@pytest.mark.asyncio
async def test_parse_metod_2(feedback):
    line = "PS=2,3;"
    parsed_states = ["2", "3"]
    await feedback._Feedback__parse_metod_2(parsed_states, line, COMMAND_FUNCTION_PRESS_SHORT, 2)
    
    assert feedback.data[COMMAND_FUNCTION_PRESS_SHORT][2][3] == 2


@pytest.mark.asyncio
async def test_parse_metod_3(feedback):
    line = "PWM=3,2,255;"
    parsed_states = ["3", "2", "255"]
    await feedback._Feedback__parse_metod_3(parsed_states, line, COMMAND_FUNCTION_PWM)

    assert feedback.data[COMMAND_FUNCTION_PWM][3][2] == "255" 


@pytest.mark.asyncio
async def test_parse_cover(feedback):
    line = "COVER=4,1,0,1,0;"
    parsed_states = ["4", "1", "0", "1", "0"]
    await feedback._Feedback__parse_cover(parsed_states, line, COMMAND_FUNCTION_COVER)

    assert feedback.data[COMMAND_FUNCTION_COVER][4] == {1: 1, 2: 0, 3: 1, 4: 0}


@pytest.mark.asyncio
async def test_parse_temp(feedback):
    line = "TEMP=5,2,25,5;"
    parsed_states = ["5", "2", "25", "5"]
    await feedback._Feedback__parse_temp(parsed_states, line)

    assert feedback.data[COMMAND_FUNCTION_TEMP][5] == {2: 25.5}


@pytest.mark.asyncio
async def test_parse_find(feedback):
    parsed_states = ["6", "123", "45"]
    await feedback._Feedback__parse_find(parsed_states)

    assert feedback.data[COMMAND_FUNCTION_FIND][6] == 123.45

@pytest.mark.asyncio
async def test_input_data_valid(feedback):
    callback = AsyncMock()
    feedback.callback = callback

    line = "IN=1,1,0,1,0,1,0,1;"
    await feedback.input_data(line)

    assert feedback.data == {
        COMMAND_FUNCTION_IN: {},
        COMMAND_FUNCTION_OUT: {},
        COMMAND_FUNCTION_PWM: {},
        COMMAND_FUNCTION_COVER: {},
        COMMAND_FUNCTION_FIND: {},
        COMMAND_FUNCTION_PONG: {},
        COMMAND_FUNCTION_TEMP: {},
        COMMAND_FUNCTION_PRESS_SHORT: {},
        COMMAND_FUNCTION_PRESS_LONG: {},
    }

@pytest.mark.asyncio
async def test_input_data_invalid_command(feedback):
    line = "INVALID=1,1,0,1;"
    await feedback.input_data(line)

    # Brak zmian w danych, ponieważ komenda jest nieznana
    assert feedback.data == {
        COMMAND_FUNCTION_IN: {},
        COMMAND_FUNCTION_OUT: {},
        COMMAND_FUNCTION_PWM: {},
        COMMAND_FUNCTION_COVER: {},
        COMMAND_FUNCTION_FIND: {},
        COMMAND_FUNCTION_PONG: {},
        COMMAND_FUNCTION_TEMP: {},
        COMMAND_FUNCTION_PRESS_SHORT: {},
        COMMAND_FUNCTION_PRESS_LONG: {},
    }




"""File with communication class for GryfApi."""

import asyncio
import logging
import re
from datetime import datetime , timedelta

from .rs232 import RS232Handler
from .tcp import TCPClientHandler
from .feedback import Feedback
from .gryf_expert import GryfExpert
from .const import (
    BAUDRATE,
    PORT,
    DriverFunctions,
    DriverActions,
)

_LOGGER = logging.getLogger(__name__)

class _GryfCommunicationApiBase():
    """Communication class for GryfApi."""
    
    _writer: RS232Handler | TCPClientHandler
    _module_count = 1
    _gryf_expert: GryfExpert
    feedback: Feedback
    _update_state_enable = True


    def __init__(
            self,
            port,
            callback=None
        ) -> None:
        """Initialise GryfCommunication."""

        self.port = port
        self._connection_task = None
        self._update_task = None
        self.feedback = Feedback(callback=callback)
        self._input_message_subscribers = []
        self._output_message_subscribers = []

        ipv4_patern = r"\b(?:\d{1,3}\.){3}\d{1,3}\b"
        serial_port_patern = r"/dev/tty\S*"

        if re.search(ipv4_patern , port):
            self._writer = TCPClientHandler(port , PORT)
        elif re.search(serial_port_patern , port):
            self._writer = RS232Handler(port , BAUDRATE)

    async def send_data(
            self,
            command
        ) -> None:
        """Send data."""

        for subscribers in self._output_message_subscribers:
            await subscribers(command)

        await self._writer.send_data(command)

    def set_callback(
            self,
            callback
        ) -> None:
        """Set callback for GryfApi data struct."""

        self.feedback.callback = callback

    def set_module_count(
            self,
            count: int,
            ) -> None:
        """Set module count for GryfApi."""

        self._module_count = count


    def subscribe_input_message(
            self,
            func,
        ) -> None:
        """Subscribe Input messages from GryfApi."""

        self._input_message_subscribers.append(func)

    def subscribe_output_message(
            self,
            func,
            ) -> None:
        """Subscribe Output messages to GryfApi."""

        self._output_message_subscribers.append(func)

    async def sending_data_to_gryf_expert(
            self,
            message: str,
        ) -> None:
        """Send data to gryf expert server."""

        if self._gryf_expert.enable == True:
            await self._gryf_expert.send_data(message)

    async def stop_connection(self) -> None:
        """Stop connection with drivers."""

        if self._connection_task:
            self._connection_task.cancel()
            try:
                await self._connection_task
            except asyncio.CancelledError:
                _LOGGER.debug("Connection task was cancelled.")
        if self._update_task:
            self._update_task.cancel()
            try:
                await self._update_task
            except asyncio.CancelledError:
                _LOGGER.debug("Update task was cancelled.")
        await self._writer.close_connection()
        _LOGGER.debug("Connection closed.")

    async def __connection_task(self) -> None:
        """Connection task."""
        try:
            while True:
                line = await self._writer.read_data()
                if line:
                    commands = line.splitlines()

                    for cmd in commands:
                        await self.feedback.input_data(cmd)
                        if self._input_message_subscribers:
                            for subscribers in self._input_message_subscribers:
                                await subscribers(cmd)
        except asyncio.CancelledError:
            _LOGGER.info("Connection task cancelled.")
            await self._writer.close_connection()
            raise
        except Exception as e:
            _LOGGER.error(f"Error in connection task: {e}")
            await self._writer.close_connection()
            raise

    async def start_connection(self) -> None:
        """Start connection with drivers."""

        await self._writer.open_connection()
        self._connection_task = asyncio.create_task(self.__connection_task())
        _LOGGER.info("Connection task started.")

    def start_update_interval(
            self,
            time: int
            ) -> None:
        """Start interval update drivers state."""

        if not self._update_task:
            self._update_task = asyncio.create_task(self.__states_update_interval(time))
            _LOGGER.info("Update interval task started.")

    async def __states_update_interval(
            self,
            time: int
        ) -> None:
        """States update interval."""
        await asyncio.sleep(1)

        try:
            while True:
                if self.feedback.data._subscription_list:
                    for key in self.feedback.data._subscription_list:
                        try:
                            value = self.feedback.data._subscription_list[key]
                            if value.inputs:
                                command = f"{DriverActions.GET_IN_STATE}={key}\n\r"
                                await self.send_data(command)
                                await asyncio.sleep(0.1)

                            if value.outputs:
                                command = f"{DriverActions.GET_OUT_STATE}={key}\n\r"
                                await self.send_data(command)
                                await asyncio.sleep(0.1)

                            if value.shutters:
                                command = f"{DriverActions.GET_SHUTTER_STATE}={key}\n\r"
                                await self.send_data(command)

                            await asyncio.sleep(5)

                        except Exception as e:
                            _LOGGER.error(f"Error updating module: {e}")
        except asyncio.CancelledError:
            _LOGGER.info("Update interval task cancelled.")
        except Exception as e:
            _LOGGER.error(f"Error in update interval: {e}")


    def available_driver(
            self, 
            id: int,
        ) -> bool:
        """Check driver available."""

        last_call = self.feedback.data.pongs.get(id , None)
        now = datetime.now()
        current_time = now.strftime("%H:%M")

        if last_call == None:
            return False

        now_dt = datetime.strptime(current_time , "%H:%M")
        last_dt = datetime.strptime(last_call , "%H:%M")

        if now_dt - last_dt < timedelta(minutes=2):
            return True
        return False



from serial_asyncio import open_serial_connection
import logging
import asyncio

_LOGGER = logging.getLogger(__name__)

class RS232Handler:
    def __init__(self, port, baudrate, reconnect_delay=5):
        self.port = port
        self.baudrate = baudrate
        self.reconnect_delay = reconnect_delay
        self.reader = None
        self.writer = None
        self._lock = asyncio.Lock()

    async def open_connection(self):
        async with self._lock:
            if self.writer and self.reader:  
                return
            try:
                self.reader, self.writer = await open_serial_connection(url=self.port, baudrate=self.baudrate)
                _LOGGER.info(f"Connection opened on port {self.port} with baudrate {self.baudrate}")

                if not self.reader or not self.writer:
                    await self.ensure_connection()
                    raise ConnectionError("Failed to initialize reader/writer.")

            except Exception as e:
                _LOGGER.error(f"Failed to open connection on port {self.port}: {e}")
                self.reader, self.writer = None, None

    async def close_connection(self):
        async with self._lock:
            if self.writer:
                try:
                    self.writer.close()
                    await self.writer.wait_closed()
                    _LOGGER.info("Connection closed successfully.")
                except Exception as e:
                    _LOGGER.error(f"Error while closing connection: {e}")
                finally:
                    self.reader, self.writer = None, None
            else:
                _LOGGER.warning("Connection was already closed or not initialized.")

    async def ensure_connection(self):
        """Ensure there is a valid connection, try to reconnect if needed."""
        attempt = 0
        if self.writer is None or self.reader is None:
            _LOGGER.warning("No connection. Attempting to reconnect...")
            while True:
                try:
                    await self.open_connection()

                    if self.writer is None or self.reader is None:
                        await self.ensure_connection()

                    return
                except Exception:
                    attempt += 1
                    _LOGGER.warning(f"Reconnect attempt {attempt + 1} failed. Retrying in {self.reconnect_delay}s...")
                    await asyncio.sleep(self.reconnect_delay)

    async def send_data(self, data):
        await self.ensure_connection()
        if self.writer:
            try:
                self.writer.write(data.encode())
                await self.writer.drain()
                _LOGGER.debug(f"Sent data: {data}")
            except Exception as e:
                _LOGGER.error(f"Error while sending data: {e}")
                error_msg = str(e)
                if "device reports readiness to read but returned no data" in error_msg:
                    _LOGGER.error("Detected device disconnection or multiple access. Closing connection.")
                    await self.close_connection()
                    await self.ensure_connection()
        else:
            _LOGGER.warning("Cannot send data: Writer is not initialized.")

    async def read_data(self):
        await self.ensure_connection()
        if self.reader:
            try:
                data = await self.reader.readuntil(b"\n")
                decoded = data.decode(errors="replace").strip()
                _LOGGER.debug(f"Read data: {decoded}")
                return decoded
            except asyncio.IncompleteReadError as e:
                _LOGGER.error(f"Incomplete read error: {e}")
                return None
            except Exception as e:
                _LOGGER.error(f"Error while reading data: {e}")
                return None
        else:
            _LOGGER.warning("Cannot read data: Reader is not initialized.")
            return None

